﻿#!/usr/bin/env python
# -*- coding: utf-8 -*-
#*******************************************************************************
#**  This file is part of HgWeb Manager.                                     ***
#**                                                                          ***
#**  Copyright (C) 2014                                                      ***
#**  CodeGNU Solutions <Licensing _AT_ CodeGNU _DOT_ com>                    ***
#**                                                                          ***
#**  This program is free software: you can redistribute it and/or modify it ***
#**  under the terms of the GNU Affero General Public License as published   ***
#**  by the Free Software Foundation, either version 3 of the License, or    ***
#**  (at your option) any later version.                                     ***
#**                                                                          ***
#**  This program is distributed in the hope that it will be useful, but     ***
#**  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
#**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
#**  See the GNU Affero General Public License for more details.             ***
#**                                                                          ***
#**  You should have received a copy of the GNU Affero General Public License***
#**  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
#*******************************************************************************

"""
 IMPORTS
"""
import manrepo
from hwm import settings


def create(repository):
	""" Creates a new repository under HgWeb.
	 @todo Check the existence of the requested repository.

	 @pre The repository to create cannot already exist.

	 @param[in] repository	The targeted repository to create.

	 @post The file system now contains a new directory in the repository path
	  that is the newly generated repository. The new repository is registered
	  in HgWeb to display and is secured to prevent accidental modification.

	 @return When the creation is successful gives true, else-wise false.
	"""
	import sys

	print "Generating repository: " + repository.StorageName
	# Create the requested repository.
	status = __generateRepository(repository)
	if 0 == status:
		print "Created repository directory."
	else:
		print >>sys.stderr, "Failed to create repository directory.", status
		return False

	# Fill out the details of about the repository.
	if __writeRepositoryDetails(repository):
		print "Wrote repository details."
	else:
		print >>sys.stderr, "Failed to write repository details."
		return False

	# Add the new repository to the web register.
	if __registerRepository(repository):
		print "Registered repository with HgWeb."
		return True
	else:
		print >>sys.stderr, "Failed to register repository with HgWeb."
		return False

def delete(repository):
	""" Removes an existing repository from HgWeb.

	 @pre The repository to remove must already exist.

	 @param[in] repository	The targeted repository to delete.

	 @post The file system no longer contains a directory in the repository
	  path that is the supplied repository. The repository is no longer
	  registered in HgWeb.

	 @return When the deletion is successful gives true, else-wise false.
	"""
	if not __doesRepositoryExist(repository):
		print >>sys.stderr, "The repository %s is not valid." % repository.StorageName
		return False

	print "Removing repository: " + repository.StorageName
	if __unregisterRepository(repository):
		print "Unregistered from HgWeb."
	else:
		print >>sys.stderr, "Failed to unregister repository from HgWeb."
		return False

	if __deleteRepository(repository):
		print "Deleted repository."
		return True
	else:
		print >>sys.stderr, "Failed to delete the repository."
		return False

def __doesRepositoryExist(repo):
	""" Checks to see if a repository exists.

	 @param[in] repo	The targeted repository for which to check.

	 @return Gives true when the repository exists, false else-wise.
	"""
	import os

	if os.path.exists(settings.RepositoryPath + os.sep + repo.StorageName) \
			and os.path.isdir(settings.RepositoryPath + os.sep + repo.StorageName) \
			and os.path.isfile(settings.RepositoryPath + os.sep + repo.StorageName + os.sep + ".hg" + os.sep + "hgrc"):
		return True
	else:
		return False

def __unregisterRepository(repo):
	""" Unregisters the repository with HgWeb.

	 @param[in] repo	The targeted repository for which to unregister.

	 @post The HgWeb configuration file does not contain the repository details
	  and is returned to a read-only state. Upon any error the standard error
	  buffer contains an error message.

	 @return When the process is successful gives true, else-wise false.
	"""
	import tempfile
	import os
	import shutil

	foundRepo=False
	with tempfile.NamedTemporaryFile() as target:
		with open(settings.HgWebPath + os.sep + 'hgweb.config', 'r') as source:
			for line in source:
				if repo.StorageName in line:
					foundRepo=True
				else:
					target.write(line)
		target.flush()
		shutil.copyfile(target.name, settings.HgWebPath + os.sep + 'hgweb.config')
	return foundRepo

def __deleteRepository(repo):
	""" Removes an existing Mercurial repository.

	 @pre The requested repository to delete must already exist.

	 @param[in] repo	The targeted repository for which to delete.

	 @post Upon any error the standard error buffer contains an error message.

	 @return When the process is successful gives true, else-wise false.
	"""
	import shutil
	import os
	import stat

	os.chmod(settings.RepositoryPath + os.sep + repo.StorageName + os.sep + ".hg" + os.sep + "hgrc", stat.S_IWRITE)
	shutil.rmtree(settings.RepositoryPath + os.sep + repo.StorageName)
	return True

def __generateRepository(repo):
	""" Initialises a new Mercurial repository.

	 @pre The requested repository to create cannot already exist.

	 @param[in] repo	The targeted repository for which to create.

	 @post Upon any error the standard error buffer contains an error message.

	 @return When the process is successful gives true, else-wise false.
	"""
	import subprocess
	import sys
	import os

	try:
		return subprocess.call([settings.HgCommand, "init", settings.RepositoryPath + os.sep + repo.StorageName])
	except OSError as e:
		print >>sys.stderr, "Initialisation error({0}): {1}".format(e.errno, e.strerror)
		return False
	return True

def __writeRepositoryDetails(repo):
	""" Writes the details of a repository to the repository configuration file.

	 @param[in] repo	The targeted repository for which to write the
					 configuration details.

	 @post The HgWeb configuration file contains the new repository details and
	  is returned to a read-only state.

	 @return When the process is successful gives true, else-wise false.
	"""
	import os
	import stat
	import sys

	config = settings.RepositoryPath + os.sep + repo.StorageName + os.sep + ".hg" + os.sep + "hgrc"
	with open(config, "a") as hgrc:
		hgrc.write("[web]\n")
		if repo.DisplayName is None:
			hgrc.write("name = " + repo.StorageName + '\n')
		else:
			hgrc.write("name = " + repo.DisplayName + '\n')
		if repo.Description is not None:
			hgrc.write("description = " + repo.Description + '\n')
		if repo.Contact is not None:
			hgrc.write("contact = " + repo.Contact + '\n')

	# Ensure the repository details isn't accidentally modified.
	hgrcMode = os.stat(config)[stat.ST_MODE]
	os.chmod(config, hgrcMode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)

	# Ensure the repository isn't accidentally deleted, moved, or blocked.
	if "win32" == sys.platform:
		status = __guardWindowsRepository(repository)
		if 0 != status:
			exit(status)
	return True

def __guardWindowsRepository(repo):
	""" Guards the supplied repository when on a Windows system.

	 @param[in] repo	The targeted repository for which to secure.

	 @return The exit status value from securing the repository. The value will
	  be zero upon success, all other values indicate failure.
	"""
	import subprocess
	import os
	import sys

	status = subprocess.call(["attrib.exe", "+H +I +S " + settings.RepositoryPath + os.sep + repo.StorageName + os.sep + ".hg"])
	if 0 == status:
		print "Secured repository store."
	else:
		print >>sys.stderr, "Failed to secure repository store.", status
	return status

def __registerRepository(repo):
	""" Registers the repository with HgWeb.

	 @param[in] repo	The targeted repository for which to register in HgWeb.

	 @post The HgWeb configuration file contains the new repository details and
	  is returned to a read-only state. Upon any error the standard error
	  buffer contains an error message.

	 @return When the process is successful gives true, else-wise false.
	"""
	import os
	import sys
	import shutil
	import stat

	try:
		with open(settings.HgWebPath + os.sep + 'hgweb.config', "a") as hgweb:
			hgweb.write('\n' + repo.StorageName + " = " + settings.RepositoryPath + os.sep + repo.StorageName)
		return True
	except IOError as e:
		print >>sys.stderr, "Web Configuration error({0}): {1}".format(e.errno, e.strerror)
		os.chmod(settings.RepositoryPath + os.sep + repo.StorageName + os.sep + ".hg" + os.sep + "hgrc", stat.S_IWRITE)
		shutil.rmtree(settings.RepositoryPath + os.sep + repo.StorageName)
		return False
